# -*- coding: utf-8 -*-
# Generated by Django 1.11.5 on 2018-02-14 20:53
from __future__ import unicode_literals

from django.db import migrations, models
import django.db.models.deletion


class Migration(migrations.Migration):

    dependencies = [
        ('apollo', '0001_initial'),
    ]

    operations = [
        migrations.CreateModel(
            name='ExternalWebhook',
            fields=[
                ('id', models.AutoField(auto_created=True, primary_key=True, serialize=False, verbose_name='ID')),
                ('url', models.CharField(max_length=255)),
                ('for_event', models.CharField(choices=[('new form submission', 'new form submission')], max_length=100)),
            ],
        ),
        migrations.AddField(
            model_name='form',
            name='allow_webhooks',
            field=models.BooleanField(default=False),
        ),
        migrations.AddField(
            model_name='externalwebhook',
            name='form',
            field=models.ForeignKey(on_delete=django.db.models.deletion.CASCADE, related_name='external_hooks', to='apollo.Form'),
        ),
    ]



from django.db.models.signals import post_save
from django.dispatch import receiver
from rest_framework.authtoken.models import Token
from .settings import apollo_settings
import django.dispatch
import logging

logger = logging.getLogger(__name__)


#####-----< Triggers >----#####
# fired when a form is submitted to the backend, but before the submission is cleaned / validated
form_submitted = django.dispatch.Signal(providing_args=["raw_data", "form_id"])

# fired after a form submission is cleaned / validated
form_submission_cleaned = django.dispatch.Signal(providing_args=["cleaned_data", "form_id", "submission_id"])

# fired on a form submission error
form_submission_error = django.dispatch.Signal(providing_args=["error", "raw_data", "form_id"])

# fired to signal a webhook to trigger
external_webhook_triggered = django.dispatch.Signal(providing_args=["hook", "data"])

# fired on a external webhook error
external_webhook_error = django.dispatch.Signal(providing_args=["error", "url", "data"])


#####-----< Listeners >----#####
@receiver(post_save)
def create_auth_token(sender, instance=None, created=False, **kwargs):
    from .models import APIUser

    if created and sender is APIUser:
        logger.debug('creating auth token for new API user')
        Token.objects.create(user=instance.auth_user)


@receiver(form_submission_cleaned)
def fire_form_submission_webhooks(sender, cleaned_data=None, form_id=None, submission_id=None, **kwargs):
    # import here to avoid circular deps
    from .models import Form, FormSubmission, ExternalWebhook
    from .api import serializers

    form = Form.objects.get(id=form_id)
    webhooks = form.external_hooks.filter(for_event=ExternalWebhook.EVENT_SUBMISSION_CREATED)

    if webhooks.exists():
        for hook in webhooks:
            data = serializers.FormSubmissionSerializer(instance=FormSubmission.objects.get(id=submission_id)).data

            external_webhook_triggered.send(sender, hook=hook, data=data)

            if not apollo_settings.EXTERNAL_HOOKS_SIGNAL_ONLY:
                logger.debug('sending submission data to hook %s' % hook)
                hook.send_data(data)



# -*- coding: utf-8 -*-

from django.test import TestCase
from smtplib import SMTPException
from apollo.lib import emails, recaptcha
import mock
import pdb


class TestSendEmail(TestCase):
    def setUp(self):
        patch_django_send_email = mock.patch('apollo.lib.emails.send_mail')
        self.mock_django_send_mail = patch_django_send_email.start()

        self.addCleanup(patch_django_send_email.stop)

    def test_calling_send_without_from_email_raises_exception(self):
        self.assertRaises(AttributeError, emails.send, None, ['something@something.com'], 'Apollo Test', 'test')

    def test_calling_send_without_to_email_raises_exception(self):
        self.assertRaises(AttributeError, emails.send, 'test@sender.com', None, 'Apollo Test', 'test')
        self.assertRaises(AttributeError, emails.send, 'test@sender.com', [], 'Apollo Test', 'test')

    def test_smtp_exception_sending_email_returns_false(self):
        self.mock_django_send_mail.side_effect = [SMTPException("error sending email")]

        self.assertFalse(emails.send(from_email='test@sender.com', to_emails=['test@receiver.com'], content='test'))

    def test_no_send_errors_returns_true(self):
        self.assertTrue(emails.send(from_email='test@sender.com', to_emails=['test@receiver.com'], content='test'))

    def test_proxies_django_send_mail(self):
        emails.send(from_email='test@sender.com', to_emails=['test@receiver.com'], subject='Apollo Test', content='test')

        self.assertIsNone(
            self.mock_django_send_mail.assert_called_with(
                'Apollo Test',
                'test',
                'test@sender.com',
                ['test@receiver.com'],
                fail_silently=False
            )
        )

    def test_handles_spanish_characters(self):
        span_chars = "años luz detrás"
        emails.send(from_email='test@sender.com', to_emails=['test@receiver.com'], subject='Apollo Test', content=span_chars)



from django.core.exceptions import ObjectDoesNotExist, ValidationError
from rest_framework import serializers
from apollo.models import Layout, LayoutField, FormField, FormFieldTemplate, Form, FormSubmission, ExternalWebhook
from apollo import signals
from apollo.settings import apollo_settings
from apollo.lib import emails, recaptcha
from apollo.exceptions import RecaptchaVerificationException
import logging
import pdb

logger = logging.getLogger(__name__)


class LayoutBlockSerializer(serializers.ModelSerializer):
    def create(self, validated_data):
        return LayoutField.objects.create(**validated_data)

    class Meta:
        model = LayoutField
        fields = ['id', 'desktop_width', 'desktop_height', 'desktop_top_left', 'mobile_width', 'mobile_height',
                  'mobile_top_left']


class LayoutSerializer(serializers.ModelSerializer):
    blocks = LayoutBlockSerializer(many=True)

    class Meta:
        model = Layout
        fields = ['id', 'name', 'max_width', 'blocks']


class FormFieldTemplateSerializer(serializers.ModelSerializer):
    class Meta:
        model = FormFieldTemplate
        fields = ['id', 'name', 'input_type', 'is_visible', 'label', 'label_position', 'placeholder', 'default_value',
                  'value_choices', 'validation_rule', 'is_submission_label']


class FieldOptionsSerializer(serializers.Serializer):
    widget_types = serializers.ListField(child=serializers.CharField())
    validation_rules = serializers.ListField(child=serializers.CharField())
    label_positions = serializers.ListField(child=serializers.CharField())


class FormFieldSerializer(serializers.ModelSerializer):
    name = serializers.CharField()
    input_type = serializers.CharField()
    is_visible = serializers.BooleanField(default=True)
    is_submission_label = serializers.BooleanField(default=False)

    def create(self, validated_data):
        field_name = validated_data.pop('name')

        template = FormFieldTemplate.objects.get(name=field_name)

        # clear out the template fields from the data, they will break things
        template_fields = ('input_type', 'is_visible', 'is_submission_label')
        [validated_data.pop(f) for f in template_fields]

        return FormField.objects.create(template=template, **validated_data)

    def update(self, instance, validated_data):
        # clear out the template fields from the data, they will break things
        template_fields = ('name', 'input_type', 'is_visible', 'is_submission_label')
        [validated_data.pop(f) for f in template_fields]

        return super(FormFieldSerializer, self).update(instance, validated_data)

    class Meta:
        model = FormField
        fields = ['id', 'name', 'input_type', 'is_visible', 'label', 'label_position', 'placeholder', 'default_value',
                  'value_choices', 'validation_rule', 'is_submission_label', 'index']


class FormSerializer(serializers.ModelSerializer):
    fields = FormFieldSerializer(many=True)

    def create(self, validated_data):
        fields = validated_data.pop('fields')

        form = Form.objects.create(**validated_data)

        self._handle_raw_fields(form, fields)

        form.save()

        return form

    def update(self, instance, validated_data):
        fields = validated_data.pop('fields')

        # check if any fields were removed from the form and delete them if so
        old_field_names = set(instance.fields.values_list('template__name', flat=True))
        new_field_names = {f['name'] for f in fields}

        to_delete = instance.fields.filter(template__name__in=old_field_names.difference(new_field_names))

        logger.debug('about to delete %s form fields' % to_delete.count())

        to_delete.delete()

        self._handle_raw_fields(instance, fields)

        return super(FormSerializer, self).update(instance, validated_data)

    def validate_include_recaptcha(self, value):
        """ recaptcha can only be included if the appropriate setting fields are set with recaptcha site key and secret """
        if value is True and not all([apollo_settings.RECAPTCHA_SITE_KEY, apollo_settings.RECAPTCHA_SECRET_KEY]):
            raise serializers.ValidationError('to use recaptcha, both a recaptcha site key and recaptcha secret need to be set')

        return value

    def _handle_raw_fields(self, form, raw_fields):
        for field_dat in raw_fields:
            try:
                field = form.fields.get(template__name=field_dat['name'])
            except ObjectDoesNotExist:
                field = None

            field_serializer = FormFieldSerializer(field, data=field_dat)

            if field_serializer.is_valid(raise_exception=True):
                field_serializer.save(form=form)

    class Meta:
        model = Form
        fields = ['id', 'name', 'redirect_url', 'include_recaptcha', 'success_message', 'submit_button_text', 'submission_url',
                  'submission_contacts', 'allow_webhooks', 'fields']


class FormSubmissionSerializer(serializers.ModelSerializer):
    cleaned_data = serializers.DictField(read_only=True)
    is_valid = serializers.BooleanField(read_only=True)
    recaptcha_token = serializers.CharField(required=False)

    def create(self, validated_data):
        request = self.context.get('request')
        form = validated_data.get('form')
        submission = form.process_submission(validated_data, request=request)

        return submission

    class Meta:
        model = FormSubmission
        fields = ['id', 'cleaned_data', 'raw_data', 'is_valid', 'created_time', 'form', 'label', 'recaptcha_token']


class ExternalWebhookSerializer(serializers.ModelSerializer):
    def validate_form(self, value):
        # if the form doesn't allow webhooks, thats an error
        if not value.allow_webhooks:
            raise ValidationError('form %s doesnt allow external webhooks. Set allow_webhooks = True and try again.' % value)

        return value

    class Meta:
        model = ExternalWebhook
        fields = ['id', 'form', 'url', 'for_event', 'service']


from django.core.mail import send_mail
from smtplib import SMTPException
import logging

logger = logging.getLogger(__name__)


def send(from_email=None, to_emails=None, subject='Apollo Email', content=None):
    """ send an
    """
    if from_email is None:
        raise AttributeError('a from_email must be set to send mail')
    elif not to_emails:
        raise AttributeError('to_emails must be set to send mail')

    logger.debug('sending email now')

    try:
        send_mail(subject, content, from_email, to_emails, fail_silently=False)
    except SMTPException as exc:
        logger.error('error sending email from %s to %s (exc: %s)' % (from_email, to_emails, str(exc)))
        return False

    return True


from django.test import TestCase
from django.core.exceptions import ValidationError
from django import forms
from apollo.models import FormFieldTemplate
from apollo.forms import SubmissionForm
from apollo.factories import FormFactory, FormFieldFactory, FormFieldTemplateFactory, FormSubmissionFactory
import mock
import pdb


class TestSubmissionForm(TestCase):
    def setUp(self):
        self.form = FormFactory.create()

        self.field_name = FormFieldFactory(
            form=self.form,
            default_value='Bob',
            validation_rule=FormFieldTemplate.RULE_REQUIRED,
            template=FormFieldTemplateFactory(name='name')
        )
        self.field_email = FormFieldFactory(
            form=self.form,
            validation_rule=FormFieldTemplate.RULE_EMAIL,
            template=FormFieldTemplateFactory(name='email')
        )
        self.field_phone = FormFieldFactory(
            form=self.form,
            validation_rule=FormFieldTemplate.RULE_OPTIONAL,
            template=FormFieldTemplateFactory(name='phone')
        )

    def _get_sform(self, fields_data, clean_data=True):
        sform = SubmissionForm(apollo_fields=[self.field_name, self.field_email, self.field_phone], data=fields_data)
        return sform

    def test_valid_data_validates(self):
        sform = self._get_sform({
            'name': 'Bobby',
            'email': 'bobby@test.com'
        })

        self.assertTrue(sform.is_valid())

        sform = self._get_sform({
            'name': 'Test',
            'email': 'bobby@test.com',
        })

        self.assertTrue(sform.is_valid())

    def test_invalid_data_doesnt_validate(self):
        sform = self._get_sform({
            'email': 'bobby@test.com'
        })

        self.assertFalse(sform.is_valid())

        sform = self._get_sform({
            'name': 'bobby',
            'email': 'bobby@'
        })

        self.assertFalse(sform.is_valid())

    def test_cleaned_data_is_correct(self):
        sform = self._get_sform({
            'name': 'Test',
            'email': 'bobby@test.com',
        })

        sform.is_valid()

        self.assertDictEqual(
            sform.cleaned_data,
            {
                'name': 'Test',
                'email': 'bobby@test.com',
                'phone': ''
            }
        )

    def test_get_widget_for_apollo_field(self):
        form = SubmissionForm()

        # text field should return text widget
        widget = form.get_widget_for_apollo_field(
            FormFieldFactory(
                form=self.form,
                template=FormFieldTemplateFactory(name='test1', input_type='TEXT')
            )
        )

        self.assertEqual(widget.input_type, 'text')

        # hidden field should return hidden widget
        widget = form.get_widget_for_apollo_field(
            FormFieldFactory(
                form=self.form,
                template=FormFieldTemplateFactory(name='test2', input_type='HIDDEN')
            )
        )

        self.assertEqual(widget, forms.HiddenInput)

        # textarea field should return textarea widget
        widget = form.get_widget_for_apollo_field(
            FormFieldFactory(
                form=self.form,
                template=FormFieldTemplateFactory(name='test3', input_type='TEXTAREA')
            )
        )

        self.assertEqual(widget, forms.Textarea)

        # select field should return select widget
        widget = form.get_widget_for_apollo_field(
            FormFieldFactory(
                form=self.form,
                template=FormFieldTemplateFactory(name='test4', input_type='SELECT')
            )
        )

        self.assertEqual(widget, forms.Select)

        # checkbox field should return checkbox widget
        widget = form.get_widget_for_apollo_field(
            FormFieldFactory(
                form=self.form,
                template=FormFieldTemplateFactory(name='test5', input_type='CHECKBOX')
            )
        )

        self.assertEqual(widget, forms.CheckboxInput)

        # richtext field should return no widget
        widget = form.get_widget_for_apollo_field(
            FormFieldFactory(
                form=self.form,
                template=FormFieldTemplateFactory(name='test6', input_type='RICHTEXT')
            )
        )

        self.assertIsNone(widget)

    def test_get_widget_attrs_for_apollo_field(self):
        form = SubmissionForm()

        attrs = form.get_widget_attrs_for_apollo_field(
            FormFieldFactory(
                form=self.form,
                placeholder=None,
                template=FormFieldTemplateFactory(name='test1')
            )
        )

        self.assertDictEqual(attrs, {})

        attrs = form.get_widget_attrs_for_apollo_field(
            FormFieldFactory(
                form=self.form,
                placeholder='foobar',
                template=FormFieldTemplateFactory(name='test2')
            )
        )

        self.assertDictEqual(attrs, {'placeholder': 'foobar'})


from django.core.exceptions import ValidationError
from django_filters.rest_framework import DjangoFilterBackend
from rest_framework import viewsets, mixins, permissions, status, views, authentication
from rest_framework.decorators import detail_route, list_route
from rest_framework.response import Response
from apollo.models import Form, FormFieldTemplate, FormField, FormSubmission, Layout
from .serializers import *
from .permissions import FormSubmissionsPermission
import pdb


class FormViewSet(viewsets.ModelViewSet):
    """
    API endpoint that allows forms to be viewed, created, updated, or deleted.

    retrieve:
    Returns a Form instance.

    Permissions Required: `null`

    list:
    Returns all Forms.

    Permissions Required: `null`

    create:
    Creates a new Form instance. The forms' Layout and Fields to include in the form are also pass-able here.

    Permissions Required: `apollo_add_form`

    update:
    Updates a Form instance.

    Permissions Required: `apollo_change_form`

    delete:
    Deletes a Form instance. Note: this will also remove all associated fields and submissions.

    Permissions Required: `apollo_remove_form`
    """
    queryset = Form.objects.all()
    serializer_class = FormSerializer
    permission_classes = (permissions.DjangoModelPermissionsOrAnonReadOnly,)
    filter_backends = (DjangoFilterBackend,)
    filter_fields = ('name',)

    @detail_route(['POST'])
    def clone(self, request, pk=None):
        form = self.get_object()
        clone = form.clone()

        return Response(FormSerializer(clone).data, status=status.HTTP_201_CREATED)

    @list_route(['POST'], url_path='import')
    def import_form(self, request, pk=None):
        imported = Form.create_from_json(request.data)
        return Response(FormSerializer(imported).data, status=status.HTTP_201_CREATED)


class FormFieldTemplateViewSet(viewsets.ModelViewSet):
    """
    API endpoint that allows FormFieldTemplates to be viewed, created, updated, or deleted.

    retrieve:
    Returns a FormFieldTemplate instance

    Permissions Required: `null`

    list:
    Returns all FormFieldTemplate instances

    Permissions Required: `null`

    create:
    Creates a new FormFieldTemplate instance.

    Permissions Required: `apollo_add_formfieldtemplate`

    update:
    Updates a FormFieldTemplate instance.

    Permissions Required: `apollo_change_formfieldtemplate`

    delete:
    Deletes a FormFieldTemplate instance.

    Permissions Required: `apollo_remove_formfieldtemplate`
    """
    queryset = FormFieldTemplate.objects.all()
    serializer_class = FormFieldTemplateSerializer
    permission_classes = (permissions.DjangoModelPermissions,)


class FormFieldViewSet(viewsets.ModelViewSet):
    """
    API endpoint that allows FormField instances to be viewed, created, updated, or deleted.

    retrieve:
    Returns a FormField instance

    Permissions Required: `null`

    list:
    Returns all FormField instances

    Permissions Required: `null`

    create:
    Creates a new FormField instance.

    Permissions Required: `apollo_add_formfield`

    update:
    Updates a FormField instance.

    Permissions Required: `apollo_change_formfield`

    delete:
    Deletes a FormField instance.

    Permissions Required: `apollo_delete_formfield`
    """
    queryset = FormField.objects.all()
    serializer_class = FormFieldSerializer
    permission_classes = (permissions.DjangoModelPermissions,)


class FormSubmissionViewSet(viewsets.ModelViewSet):
    """
    API endpoint that allows Submission instances to be viewed, created, updated, or deleted.

    retrieve:
    Returns a FormSubmission instance

    Permissions Required: `is_staff = True`

    list:
    Returns all FormSubmission instances

    Permissions Required: `is_staff = True`

    create:
    Creates a new FormSubmission instance.

    Permissions Required: `null`

    update:
    Updates a FormSubmission instance.

    Permissions Required: `is_staff = True`

    delete:
    Deletes a FormSubmission instance.

    Permissions Required: `is_staff = True`
    """
    queryset = FormSubmission.objects.all()
    serializer_class = FormSubmissionSerializer
    authentication_classes = (authentication.SessionAuthentication, authentication.BasicAuthentication, authentication.TokenAuthentication,)
    permission_classes = (FormSubmissionsPermission,)

    filter_backends = (DjangoFilterBackend,)
    filter_fields = ('form_id',)

    def create(self, request, *args, **kwargs):
        try:
            return super(FormSubmissionViewSet, self).create(request, *args, **kwargs)
        except ValidationError as exc:
            return Response({'error': exc.message}, status=status.HTTP_400_BAD_REQUEST)


class LayoutViewSet(viewsets.ModelViewSet):
    """
    API endpoint that allows Layout instances to be viewed, created, updated, or deleted.

    retrieve:
    Returns a Layout instance

    Permissions Required: `null`

    list:
    Returns all Layout instances

    Permissions Required: `null`

    create:
    Creates a new Layout instance.

    Permissions Required: `apollo_add_layout`

    update:
    Updates a Layout instance.

    Permissions Required: `apollo_change_layout`

    delete:
    Deletes a Layout instance.

    Permissions Required: `apollo_delete_layout`
    """
    queryset = Layout.objects.all()
    serializer_class = LayoutSerializer
    permission_classes = (permissions.DjangoModelPermissions,)


class ExternalWebhookViewSet(
    mixins.CreateModelMixin,
    mixins.DestroyModelMixin,
    viewsets.GenericViewSet
):
    """
    API endpoint that allows ExternalWebhook instances to be created and deleted.

    create:
    Creates a new ExternalWebhook instance.

    Permissions Required: `apollo_add_externalwebhook`

    delete:
    Deletes a ExternalWebhook instance

    Permissions Required: `apollo_delete_externalwebhook`
    """
    queryset = ExternalWebhook.objects.all()
    serializer_class = ExternalWebhookSerializer
    permission_classes = (permissions.DjangoModelPermissions,)
    authentication_classes = (authentication.SessionAuthentication, authentication.BasicAuthentication, authentication.TokenAuthentication,)


class FormFieldOptionsAPIView(views.APIView):
    """
    API endpoint that allows FormFieldOptions to be viewed.

    list:
    Returns all FormFieldOptions, including Widget Types (e.g. 'TEXT', 'SELECT'), Validation Rules (e.g. 'REQUIRED', 'OPTIONAL'),
    and Label Positions (e.g. 'TOP', 'LEFT')

    Permissions Required: `null`
    """
    def get(self, request, *args, **kwargs):
        serializer = FieldOptionsSerializer({
            'widget_types': [_[0] for _ in FormFieldTemplate.FIELD_TYPES],
            'validation_rules': [_[0] for _ in FormFieldTemplate.VALIDATION_RULES],
            'label_positions': [_[0] for _ in FormFieldTemplate.LABEL_POSITIONS]
        })

        return Response(serializer.data)



from importlib import import_module
from apollo.settings import apollo_settings


urls = import_module('apollo.api.%s.urls' % apollo_settings.API_VERSION)
serializers = import_module('apollo.api.%s.serializers' % apollo_settings.API_VERSION)


from django.contrib import admin
from apollo import models


class APIUserAdmin(admin.ModelAdmin):
    fields = ['service_name', 'auth_user']


class ExternalWebhookAdmin(admin.ModelAdmin):
    fields = ['form', 'url', 'for_event', 'service']
    list_display = ['form', 'url', 'service']


admin.site.register(models.APIUser, APIUserAdmin)
admin.site.register(models.ExternalWebhook, ExternalWebhookAdmin)


import requests
from apollo.exceptions import RecaptchaVerificationException
import logging

logger = logging.getLogger(__name__)


def verify_token(token, recaptcha_secret_key=None):
    """ takes a token as generated by a frontend verification request (see https://developers.google.com/recaptcha/docs/v3
    for reference) and verifies the token against Googles' services
    :param token: `str` token as given by the frontend
    :param recaptcha_secret_key: `str` secret key for the recaptcha service
    :return: `bool` True if the verification succeeded, False otherwise
    :raises: `apollo.exceptions.RecaptchaVerificationException` if verification returns an error code
    """
    url = 'https://www.google.com/recaptcha/api/siteverify'
    data = {
        'secret': recaptcha_secret_key,
        'response': token
    }

    resp = requests.post(url, data=data)

    if not resp.ok:
        logger.error('got bad response from recaptcha API: %s' % resp.content)
        raise RecaptchaVerificationException('got error response from recaptcha verification (status = %s)' % resp.status_code)

    resp_json = resp.json()
    logger.debug('got response from recaptcha verification: JSON = %s' % resp_json)

    if resp_json.get('error-codes'):
        raise RecaptchaVerificationException('error codes in recaptcha verification: %s' % resp_json.get('error-codes'))

    return resp_json.get('success')


from __future__ import unicode_literals

from django.apps import AppConfig


class ApolloConfig(AppConfig):
    name = 'apollo'

    def ready(self):
        super(ApolloConfig, self).ready()
        import apollo.signals



# -*- coding: utf-8 -*-
# Generated by Django 1.11.29 on 2022-08-16 20:13
from __future__ import unicode_literals

from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [
        ('apollo', '0003_externalwebhook_service'),
    ]

    operations = [
        migrations.AddField(
            model_name='form',
            name='include_recaptcha',
            field=models.BooleanField(default=False),
        ),
        migrations.AlterField(
            model_name='formfield',
            name='validation_rule',
            field=models.CharField(blank=True, choices=[('OPTIONAL', 'OPTIONAL'), ('REQUIRED', 'REQUIRED'), ('EMAIL', 'EMAIL'), ('PHONE_NUMBER', 'PHONE_NUMBER')], default='REQUIRED', max_length=50, null=True),
        ),
        migrations.AlterField(
            model_name='formfieldtemplate',
            name='validation_rule',
            field=models.CharField(blank=True, choices=[('OPTIONAL', 'OPTIONAL'), ('REQUIRED', 'REQUIRED'), ('EMAIL', 'EMAIL'), ('PHONE_NUMBER', 'PHONE_NUMBER')], default='REQUIRED', max_length=50, null=True),
        ),
    ]



from rest_framework import permissions


class FormSubmissionsPermission(permissions.DjangoModelPermissions):
    """ form submissions may be viewed by staff only, but are open for creation by anyone """
    message = 'Only users with perms may view form submissions'

    def has_permission(self, request, view):
        # anyone can create a form submission
        if request.method == 'POST':
            return True

        has_modify_perms = super(FormSubmissionsPermission, self).has_permission(request, view)
        has_read_perms = request.user.has_perm('apollo.can_view_submissions')

        return has_modify_perms if request.method != 'GET' else has_read_perms



from django.test import TestCase, Client, RequestFactory
from contextlib import contextmanager
from rest_framework.test import APIClient
from django.contrib.auth.models import User
from django.core.exceptions import ValidationError, ObjectDoesNotExist
from django.contrib.auth.models import Permission
from rest_framework.authtoken.models import Token
from apollo.factories import FormFactory, FormFieldFactory, FormFieldTemplateFactory, FormSubmissionFactory, APIUserFactory
from apollo.models import Form, FormFieldTemplate, ExternalWebhook
from apollo.settings import apollo_settings
from apollo.api.v1.serializers import FormSerializer, FormFieldSerializer
from apollo import signals
import os
import factory
import mock
import json
import pdb
import logging
import re
import copy

logger = logging.getLogger(__name__)


class APITestCase(TestCase):
    base_path = '/api'
    version = 'v1'
    should_login = True
    user_username = 'tester'
    user_password = 'mytestapi'

    def setUp(self):
        self.client = APIClient()

        self.user = User.objects.create_user(username=self.user_username, email='test@test.com', password=self.user_password)

        if self.should_login:
            self.login_client()

        patch_send_email = mock.patch('apollo.lib.emails.send', return_value=True)
        self.mock_send_email = patch_send_email.start()

        self.addCleanup(patch_send_email.stop)

    def add_permission(self, codename):
        perm = Permission.objects.get(codename=codename)
        self.user.user_permissions.add(perm)

    def remove_permission(self, codename):
        perm = Permission.objects.get(codename=codename)
        self.user.user_permissions.remove(perm)

    def abs_path(self, path):
        return '%s/%s%s' % (self.base_path, self.version, path)

    def login_client(self, auth_params=None):
        auth_params = auth_params or {'username': 'tester', 'password': 'mytestapi'}

        self.client.login(**auth_params)

    def post(self, path, params=None):
        params = params or {}

        return self.client.post(self.abs_path(path), params, format='json')

    def put(self, path, params=None):
        params = params or {}
        return self.client.put(self.abs_path(path), data=params, format='json')

    def delete(self, path, params=None):
        params = params or {}
        return self.client.delete(self.abs_path(path), data=params, format='json')

    def get(self, path, params=None):
        return self.client.get(self.abs_path(path), data=params, format='json')


#####-----< Forms API >----#####
#####-----< Forms API >----#####
class FormsAPITestCase(APITestCase):
    def setUp(self):
        super(FormsAPITestCase, self).setUp()

        self._first_name_template = FormFieldTemplateFactory(name='first_name', input_type='TEXT', is_visible=True)
        self._last_name_template = FormFieldTemplateFactory(name='last_name', input_type='TEXT', is_visible=True)


class TestCreateForm(FormsAPITestCase):
    def setUp(self):
        super(TestCreateForm, self).setUp()

        self.form_data = {
            'name': 'Test Form',
            'redirect_url': 'http://test.com',
            'submit_button_text': 'Submit',
            'submission_url': '/submissions',
            'fields': [
                {
                    'name': 'first_name',
                    'input_type': 'TEXT',
                    'is_visible': True,
                    'placeholder': 'First Name'
                },
                {
                    'name': 'last_name',
                    'input_type': 'TEXT',
                    'is_visible': True,
                    'placeholder': 'Last Name'
                }
            ]
        }

        self.add_permission('add_form')

    def test_create_form_with_basic_data(self):
        resp = self.post('/forms', self.form_data)

        self.assertEqual(resp.status_code, 201)

        form = Form.objects.get(id=resp.json()['id'])

        self.assertEqual(form.name, self.form_data['name'])
        self.assertEqual(form.redirect_url, self.form_data['redirect_url'])
        self.assertEqual(form.submit_button_text, self.form_data['submit_button_text'])
        self.assertEqual(form.submission_url, self.form_data['submission_url'])

        for field in self.form_data['fields']:
            try:
                ff = form.fields.get(template__name=field['name'])
            except ObjectDoesNotExist:
                self.fail('%s was never created for form %s' % (field['name'], form))

            for attr, val in field.items():
                self.assertEqual(getattr(ff, attr), val)

    def test_create_form_perms_required_to_create_form(self):
        self.remove_permission('add_form')

        resp = self.post('/forms', self.form_data)

        self.assertEqual(resp.status_code, 403)


class TestUpdateForm(FormsAPITestCase):
    def setUp(self):
        super(TestUpdateForm, self).setUp()

        self.form = FormFactory(name='Test Form', redirect_url='http://test.com', submit_button_text='Submit')
        self.first_name_field = FormFieldFactory(template=self._first_name_template, placeholder='First Name', form=self.form)
        self.last_name_field = FormFieldFactory(template=self._last_name_template, placeholder='Last Name', form=self.form)

        self.form_data_template = {
            'name': self.form.name,
            'redirect_url': self.form.redirect_url,
            'submit_button_text': self.form.submit_button_text,
            'submission_url': self.form.submission_url,
            'fields': [
                {
                    'name': f.name,
                    'input_type': f.input_type,
                    'is_visible': f.is_visible,
                    'placeholder': f.placeholder
                } for f in (self.first_name_field, self.last_name_field)
            ]
        }

        self.add_permission('change_form')

    def test_update_form_meta(self):
        update_data = copy.deepcopy(self.form_data_template)

        update_data.update({
            'name': 'Test Form 2',
            'redirect_url': 'https://test.com'
        })

        resp = self.put('/forms/%s' % self.form.id, update_data)

        self.assertEqual(resp.status_code, 200)

        self.form.refresh_from_db()

        self.assertEqual(self.form.name, update_data['name'])
        self.assertEqual(self.form.redirect_url, update_data['redirect_url'])
        self.assertEqual(self.form.submit_button_text, self.form_data_template['submit_button_text'])
        self.assertEqual(self.form.submission_url, self.form_data_template['submission_url'])

    def test_update_form_fields(self):
        update_data = copy.deepcopy(self.form_data_template)

        update_data.update({
            'fields': [
                {
                    'name': 'first_name',
                    'input_type': 'TEXT',
                    'is_visible': True,
                    'placeholder': 'Nickname'
                },
                {
                    'name': 'last_name',
                    'input_type': 'TEXT',
                    'is_visible': True,
                    'placeholder': 'Last'
                }
            ]
        })

        self.put('/forms/%s' % self.form.id, update_data)

        self.first_name_field.refresh_from_db()
        self.last_name_field.refresh_from_db()

        self.assertEqual(self.first_name_field.placeholder, 'Nickname')
        self.assertEqual(self.last_name_field.placeholder, 'Last')

    def test_remove_form_field(self):
        update_data = copy.deepcopy(self.form_data_template)

        update_data.update({
            'fields': [
                {
                    'name': 'first_name',
                    'input_type': 'TEXT',
                    'is_visible': True,
                    'placeholder': 'First Name'
                }
            ]
        })

        self.put('/forms/%s' % self.form.id, update_data)

        self.form.refresh_from_db()

        self.assertEqual(self.form.fields.count(), 1)
        self.assertRaises(ObjectDoesNotExist, self.last_name_field.refresh_from_db)

    def test_add_form_field(self):
        FormFieldTemplateFactory(name='email', input_type='EMAIL', is_visible=True)

        update_data = copy.deepcopy(self.form_data_template)

        update_data.update({
            'fields': [
                {
                    'name': 'first_name',
                    'input_type': 'TEXT',
                    'is_visible': True,
                    'placeholder': 'First Name'
                },
                {
                    'name': 'last_name',
                    'input_type': 'TEXT',
                    'is_visible': True,
                    'placeholder': 'Last Name'
                },
                {
                    'name': 'email',
                    'input_type': 'EMAIL',
                    'is_visible': True,
                    'placeholder': 'Email'
                },
            ]
        })

        self.put('/forms/%s' % self.form.id, update_data)

        self.form.refresh_from_db()

        self.assertEqual(self.form.fields.count(), 3)
        self.assertListEqual(
            list(self.form.fields.values_list('template__name', flat=True).order_by('template__name')),
            sorted(['first_name', 'last_name', 'email'])
        )

    def test_change_permissions_required_to_change_form(self):
        self.remove_permission('change_form')

        update_data = copy.deepcopy(self.form_data_template)

        update_data.update({
            'name': 'Test Form 2',
            'redirect_url': 'https://test.com'
        })

        resp = self.put('/forms/%s' % self.form.id, update_data)

        self.assertEqual(resp.status_code, 403)


class TestRetrieveForm(FormsAPITestCase):
    def setUp(self):
        super(TestRetrieveForm, self).setUp()

        self.form = FormFactory()
        self.first_name_field = FormFieldFactory(validation_rule='REQUIRED', template=self._first_name_template, form=self.form)
        self.last_name_field = FormFieldFactory(validation_rule='REQUIRED', template=self._last_name_template, form=self.form)

    def test_retrieve_basic_form_data(self):
        resp = self.get('/forms/%s' % self.form.id)

        form_data = resp.json()

        self.assertEqual(form_data['name'], self.form.name)
        self.assertEqual(form_data['submission_url'], self.form.submission_url)
        self.assertEqual(form_data['redirect_url'], self.form.redirect_url)

        for field in self.form.fields.all():
            dat_field = [f for f in form_data['fields'] if f['name'] == field.name][0]

            self.assertEqual(dat_field['name'], field.name)
            self.assertEqual(dat_field['input_type'], field.input_type)
            self.assertEqual(dat_field['validation_rule'], field.validation_rule)

    def test_no_perms_required_to_retrieve_form_data(self):
        self.client.logout()

        resp = self.get('/forms/%s' % self.form.id)

        self.assertEqual(resp.status_code, 200)


class TestCloneForm(FormsAPITestCase):
    def setUp(self):
        super(TestCloneForm, self).setUp()

        self.form = FormFactory()

        self.add_permission('add_form')

    @mock.patch('apollo.models.Form.clone')
    def test_clone_form_calls_clone_method(self, mock_clone):
        resp = self.post('/forms/%s/clone' % self.form.id)

        self.assertEqual(mock_clone.call_count, 1)

    def test_clone_form_returns_clone(self):
        resp = self.post('/forms/%s/clone' % self.form.id)

        self.assertEqual(resp.status_code, 201)

        self.assertGreater(resp.json()['id'], self.form.id)

    def test_clone_form_requires_add_form_permission(self):
        self.remove_permission('add_form')

        resp = self.post('/forms/%s/clone' % self.form.id)

        self.assertEqual(resp.status_code, 403)


#####-----< FormFieldTemplate API >----#####
#####-----< FormFieldTemplate API >----#####
class TestCreateFormFieldTemplate(APITestCase):
    def setUp(self):
        super(TestCreateFormFieldTemplate, self).setUp()

        self.field_template_data = {
            'name': 'first_name',
            'input_type': 'TEXT',
            'is_visible': True,
            'label': 'First Name',
            'placeholder': 'First Name',
            'validation_rule': 'REQUIRED'
        }

        self.add_permission('add_formfieldtemplate')

    def test_create_basic_form_field_template(self):
        resp = self.post('/field-templates', self.field_template_data)

        self.assertEqual(resp.status_code, 201)

        resp_data = resp.json()

        created_template = FormFieldTemplate.objects.get(id=resp_data['id'])

        for k, v in self.field_template_data.items():
            self.assertEqual(getattr(created_template, k), v)

    def test_creating_form_field_template_requires_create_form_field_template_permission(self):
        self.remove_permission('add_formfieldtemplate')

        resp = self.post('/field-templates', self.field_template_data)

        self.assertEqual(resp.status_code, 403)


class TestUpdateFormFieldTemplate(APITestCase):
    def setUp(self):
        super(TestUpdateFormFieldTemplate, self).setUp()

        self.field_template = FormFieldTemplateFactory(name='email', input_type='EMAIL', placeholder='Email', label='Email')

        self.field_template_data = {
            'name': self.field_template.name,
            'input_type': self.field_template.input_type,
            'is_visible': self.field_template.is_visible,
            'label': self.field_template.label,
            'placeholder': self.field_template.placeholder,
            'validation_rule': self.field_template.validation_rule
        }

        self.add_permission('change_formfieldtemplate')

    def test_changing_field_value(self):
        update_data = copy.deepcopy(self.field_template_data)

        update_data.update({
            'input_type': 'TEXT',
            'label': None,
        })

        resp = self.put('/field-templates/%s' % self.field_template.id, update_data)

        self.assertEqual(resp.status_code, 200)

        self.field_template.refresh_from_db()

        self.assertEqual(self.field_template.input_type, update_data['input_type'])
        self.assertEqual(self.field_template.label, update_data['label'])
        self.assertEqual(self.field_template.name, self.field_template_data['name'], 'the field name should not have changed')

    def test_change_form_field_template_requires_change_form_field_template_permission(self):
        self.remove_permission('change_formfieldtemplate')

        resp = self.post('/field-templates', self.field_template_data)

        self.assertEqual(resp.status_code, 403)


#####-----< FormField API >----#####
#####-----< FormField API >----#####
## TODO: FILL ME IN ##


#####-----< Submissions API >----#####
#####-----< Submissions API >----#####
class TestCreateFormSubmission(APITestCase):
    def setUp(self):
        super(TestCreateFormSubmission, self).setUp()

        # set up DB constructs
        self.form = FormFactory(id=1, name='Test Form')
        FormFieldFactory(form=self.form, template=FormFieldTemplateFactory(name='test'))

        self.submission_data = {
            'form': 1,
            'raw_data': [
                {
                    'id': 1,
                    'name': 'test',
                    'input_type': 'TEXT',
                    'is_visible': True,
                    'label': 'Test',
                    'validation_rule': 'OPTIONAL',
                    'value': 'Testy',
                }
            ]
        }

    @factory.django.mute_signals(signals.form_submission_cleaned)
    def test_returns_status_of_success_on_success(self):
        with mock.patch('apollo.models.FormSubmission.clean_data', return_value={}) as mock_clean_data:
            resp = self.post('/submissions', self.submission_data)

            self.assertEqual(mock_clean_data.call_count, 1)
            self.assertEqual(resp.status_code, 201)
            self.assertGreater(resp.json()['id'], 0)

    @factory.django.mute_signals(signals.form_submission_cleaned)
    def test_returns_status_of_error_on_value_error(self):
        with mock.patch('apollo.models.FormSubmission.clean_data', side_effect=ValueError) as mock_clean_data:
            resp = self.post('/submissions', self.submission_data)

            self.assertEqual(resp.status_code, 400)
            self.assertEqual(mock_clean_data.call_count, 1)
            self.assertEqual(resp.json()['error'], 'got error cleaning submission data')

    def test_returns_status_of_error_on_validation_error(self):
        with mock.patch('apollo.models.FormSubmission.clean_data', side_effect=ValidationError('error validating')) as mock_clean_data:
            resp = self.post('/submissions', self.submission_data)

            self.assertEqual(resp.status_code, 400)
            self.assertEqual(mock_clean_data.call_count, 1)
            self.assertEqual(resp.json()['error'], 'got error validating submission data')

    @factory.django.mute_signals(signals.form_submission_cleaned)
    def test_creates_new_submission_instance_when_clean_succeeds(self):
        resp = self.post('/submissions', self.submission_data)

        created_submission = self.form.submissions.first()

        self.assertDictEqual(
            created_submission.raw_data[0],
            self.submission_data['raw_data'][0]
        )
        self.assertDictEqual(created_submission.cleaned_data, {'test': 'Testy'})

    def test_still_creates_submission_when_clean_fails(self):
        with mock.patch('apollo.models.FormSubmission.clean_data', side_effect=ValidationError('error validating')) as mock_clean_data:
            resp = self.post('/submissions', self.submission_data)

            self.assertEqual(resp.status_code, 400)
            self.assertEqual(self.form.submissions.count(), 1)
            self.assertIsNone(self.form.submissions.first().cleaned_data)

    @factory.django.mute_signals(signals.form_submission_cleaned)
    def test_sends_email_to_submission_contacts_when_set(self):
        resp = self.post('/submissions', self.submission_data)

        self.assertEqual(self.mock_send_email.call_count, 0, "shouldnt send email when form has no submission contacts")

        # now try with submission contacts
        self.form.submission_contacts = ['bobby@classaction.com']
        self.form.save()

        self.post('/submissions', self.submission_data)

        self.assertEqual(self.mock_send_email.call_count, 1)
        self.assertIsNone(
            self.mock_send_email.assert_called_with(
                from_email=apollo_settings.SUBMISSION_EMAIL_FROM,
                to_emails=[u'bobby@classaction.com'],
                subject='Apollo: New Form Submission',
                content='new form submission, data = %s' % {u'test': u'Testy'}
            )
        )

    @factory.django.mute_signals(signals.form_submission_cleaned)
    def test_no_perms_required_to_create_submission(self):
        self.client.logout()

        resp = self.post('/submissions', self.submission_data)

        self.assertEqual(resp.status_code, 201)


class TestListFormSubmissions(FormsAPITestCase):
    def setUp(self):
        super(TestListFormSubmissions, self).setUp()

        self.add_permission('can_view_submissions')
        self.add_permission('change_formsubmission')
        self.add_permission('delete_formsubmission')

        self.form = FormFactory(name='Test Form')
        self.first_name_field = FormFieldFactory(template=self._first_name_template, form=self.form, placeholder='First Name')
        self.last_name_field = FormFieldFactory(template=self._last_name_template, form=self.form, placeholder='Last Name')

    def test_basic_list_submissions(self):
        submissions = []
        for field in (self.first_name_field, self.last_name_field):
            submissions.append(FormSubmissionFactory(form=self.form, raw_data={
                'id': field.id,
                'name': field.name,
                'input_type': field.input_type,
                'is_visible': field.is_visible,
                'value': factory.Faker(field.name).generate({})
            }))

        resp = self.get('/submissions')

        self.assertEqual(resp.status_code, 200)

        self.assertListEqual(sorted([s['id'] for s in resp.json()]), sorted([s.id for s in submissions]))

    def test_filtering_submissions_by_form(self):
        other_form = FormFactory(name='Some Form')
        other_submissions = FormSubmissionFactory(form=other_form, raw_data={})

        submissions = []
        for field in (self.first_name_field, self.last_name_field):
            submissions.append(FormSubmissionFactory(form=self.form, raw_data={
                'id': field.id,
                'name': field.name,
                'input_type': field.input_type,
                'is_visible': field.is_visible,
                'value': factory.Faker(field.name).generate({})
            }))

        resp = self.get('/submissions', params={'form_id': self.form.id})

        self.assertListEqual(sorted([s['id'] for s in resp.json()]), sorted([s.id for s in self.form.submissions.all()]))

    def test_can_view_submissions_permission_required_to_view_submissions(self):
        self.remove_permission('can_view_submissions')

        resp = self.get('/submissions')

        self.assertEqual(resp.status_code, 403)

    def test_can_change_submission_permissions_required_to_update_submission(self):
        submission = FormSubmissionFactory(
            form=self.form,
            raw_data={'fields': [{'name': 'first_name', 'value': 'bobby'}, {'name': 'last_name', 'value': 'stein'}]}
        )

        self.remove_permission('change_formsubmission')

        resp = self.put('/submissions/%s' % submission.id, params={'raw_data': { 'is_valid': False }, 'form': self.form.id})

        self.assertEqual(resp.status_code, 403)

    def test_can_delete_form_submission_permissions_required_to_delete_submission(self):
        submission = FormSubmissionFactory(
            form=self.form,
            raw_data={'fields': [{'name': 'first_name', 'value': 'bobby'}, {'name': 'last_name', 'value': 'stein'}]}
        )

        self.remove_permission('delete_formsubmission')

        resp = self.delete('/submissions/%s' % submission.id)

        self.assertEqual(resp.status_code, 403)


class TestTokenAuthentication(FormsAPITestCase):
    should_login = False

    def setUp(self):
        super(TestTokenAuthentication, self).setUp()

        self.api_user = APIUserFactory(auth_user=self.user)

        self.add_permission('can_view_submissions')

        self.form = FormFactory(name='Test Form')
        self.first_name_field = FormFieldFactory(template=self._first_name_template, form=self.form, placeholder='First Name')
        self.last_name_field = FormFieldFactory(template=self._last_name_template, form=self.form, placeholder='Last Name')

    def test_api_user_with_auth_token_can_read_form_submissions(self):
        submissions = []

        for field in (self.first_name_field, self.last_name_field):
            submissions.append(FormSubmissionFactory(form=self.form, raw_data={
                'id': field.id,
                'name': field.name,
                'input_type': field.input_type,
                'is_visible': field.is_visible,
                'value': factory.Faker(field.name).generate({})
            }))

        self.client.credentials(HTTP_AUTHORIZATION='Token ' + Token.objects.get(user=self.user).key)
        resp = self.get('/submissions')

        self.assertEqual(resp.status_code, 200)

        self.assertListEqual(sorted([s['id'] for s in resp.json()]), sorted([s.id for s in submissions]))


class TestRetrieveAuthToken(FormsAPITestCase):
    should_login = False

    def setUp(self):
        super(TestRetrieveAuthToken, self).setUp()

        self.api_user = APIUserFactory(auth_user=self.user)

    def test_passing_incorrect_creds_returns_400(self):
        resp = self.post('/auth-token', {'username': 'badusername', 'password': 'badpass'})

        self.assertEqual(resp.status_code, 400)

    def test_passing_correct_creds_returns_200_and_token(self):
        resp = self.post('/auth-token', {'username': self.user_username, 'password': self.user_password})

        self.assertEqual(resp.status_code, 200)
        self.assertDictEqual(resp.json(), {'token': Token.objects.get(user=self.user).key})


class TestExternalWebhookRegistration(FormsAPITestCase):
    def setUp(self):
        super(TestExternalWebhookRegistration, self).setUp()

        self.add_permission('add_externalwebhook')
        self.add_permission('delete_externalwebhook')

    def test_allows_registration_to_form_with_allow_webhooks(self):
        form = FormFactory(allow_webhooks=True)

        data = {'service': 'zapier', 'for_event': ExternalWebhook.EVENT_SUBMISSION_CREATED, 'url': 'https://test.com', 'form': form.id}

        resp = self.post('/hooks', data)

        self.assertEqual(resp.status_code, 201)

        hook = ExternalWebhook.objects.get(id=resp.json()['id'])

        self.assertEqual(hook.service, data['service'])
        self.assertEqual(hook.for_event, data['for_event'])
        self.assertEqual(hook.url, data['url'])

    def test_doesnt_allow_registration_to_form_without_allow_webhooks(self):
        form = FormFactory(allow_webhooks=False)

        data = {'service': 'zapier', 'for_event': ExternalWebhook.EVENT_SUBMISSION_CREATED, 'url': 'https://test.com', 'form': form.id}

        resp = self.post('/hooks', data)

        self.assertEqual(resp.status_code, 400)
        print(resp.content)

    def test_only_allows_webhook_creation_and_deletion(self):
        pass


# -*- coding: utf-8 -*-
# Generated by Django 1.11.5 on 2018-02-14 21:37
from __future__ import unicode_literals

from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [
        ('apollo', '0002_auto_20180214_2053'),
    ]

    operations = [
        migrations.AddField(
            model_name='externalwebhook',
            name='service',
            field=models.CharField(default='zapier', max_length=100),
            preserve_default=False,
        ),
    ]



from django.test import TestCase
from django.core.exceptions import ValidationError
from django.test.client import RequestFactory
from apollo.models import FormFieldTemplate, ExternalWebhook, Form
from apollo.factories import FormFactory, FormFieldFactory, FormFieldTemplateFactory, FormSubmissionFactory, ExternalWebhookFactory
import json
import pdb
import mock
import requests
import responses


class MockResponse(responses.Response):
    pass


#####-----< Form >----#####
class FormTestCase(TestCase):
    def setUp(self):
        self.form = FormFactory.create()

    def test_create_from_json_with_valid_data(self):
        valid_data = {
            "id": 2,
            "name": "Test Form 2",
            "redirect_url": None,
            "include_recaptcha": False,
            "success_message": "Success",
            "submit_button_text": "Submit",
            "submission_url": None,
            "submission_contacts": None,
            "allow_webhooks": False,
            "fields": [{
                "id":  3,
                "name": "last_name",
                "input_type": "TEXT",
                "is_visible": True,
                "label": "Last Name",
                "label_position": "TOP",
                "placeholder": "Last Name",
                "default_value": None,
                "value_choices": None,
                "validation_rule": "OPTIONAL",
                "is_submission_label": False,
                "index": 0
            }, {
                "id": 4,
                "name": "first_name",
                "input_type": "TEXT",
                "is_visible": True,
                "label": "First Name",
                "label_position": "TOP",
                "placeholder": "First Name",
                "default_value": None,
                "value_choices": None,
                "validation_rule": "OPTIONAL",
                "is_submission_label": False,
                "index": 0
            }]
        }

        form = Form.create_from_json(valid_data)

        self.assertEqual(form.fields.count(), 2)
        self.assertEqual(FormFieldTemplate.objects.count(), 2)
        self.assertEqual(form.fields.get(template__name='last_name').label, 'Last Name')
        self.assertEqual(form.fields.get(template__name='first_name').label, 'First Name')

        # if we modify the form name and perform the same op, we should have another, new form. Verify it doesn't o
        # override the existing form instead
        valid_data['name'] = 'Test Form 3'
        new_form = Form.create_from_json(valid_data)

        self.assertNotEqual(new_form.id, 2)
        self.assertEqual(new_form.fields.count(), 2)
        self.assertEqual(FormFieldTemplate.objects.count(), 2)

    def test_create_from_json_with_invalid_data(self):
        valid_data = {
            "id": 2,
            "name": "Test Form 2",
            "redirect_url": None,
            "include_recaptcha": False,
            "success_message": "Success",
            "submit_button_text": "Submit",
            "submission_url": None,
            "submission_contacts": None,
            "allow_webhooks": False,
            "fields": [{
                "id": 3,
                "name": "last_name",
                "input_type": "TEXT",
                "is_visible": True,
                "label": "Last Name",
                "label_position": "TOP",
                "placeholder": "Last Name",
                "default_value": None,
                "value_choices": None,
                "validation_rule": "OPTIONAL",
                "is_submission_label": False,
                "index": 0
            }, {
                "id": 4,
                "name": "first_name",
                "input_type": "TEXT",
                "is_visible": True,
                "label": "First Name",
                "label_position": "TOP",
                "placeholder": "First Name",
                "default_value": None,
                "value_choices": None,
                "validation_rule": "OPTIONAL",
                "is_submission_label": False,
                "index": 0
            }]
        }

        form = Form.create_from_json(valid_data)

        # invalid because duplicate form
        invalid_data = {
            "id": 2,
            "name": "Test Form 2",
            "redirect_url": None,
            "include_recaptcha": False,
            "success_message": "Success",
            "submit_button_text": "Submit",
            "submission_url": None,
            "submission_contacts": None,
            "allow_webhooks": False,
            "fields": [{
                "id": 3,
                "name": "last_name",
                "input_type": "TEXT",
                "is_visible": True,
                "label": "Last Name",
                "label_position": "TOP",
                "placeholder": "Last Name",
                "default_value": None,
                "value_choices": None,
                "validation_rule": "OPTIONAL",
                "is_submission_label": False,
                "index": 0
            }, {
                "id": 4,
                "name": "first_name",
                "input_type": "TEXT",
                "is_visible": True,
                "label": "First Name",
                "label_position": "TOP",
                "placeholder": "First Name",
                "default_value": None,
                "value_choices": None,
                "validation_rule": "OPTIONAL",
                "is_submission_label": False,
                "index": 0
            }]
        }

        try:
            form = Form.create_from_json(invalid_data)
            self.assertTrue(False, 'it should have thrown an error with invalid data')
        except ValidationError:
            self.assertTrue(True)


#####-----< Form Submissions >----#####
class FormSubmissionTestCase(TestCase):
    default_raw_data = [
        {'name': 'name', 'value': 'bob', 'id': 1, 'validation_rule': 'REQUIRED'},
        {'name': 'email', 'value': 'bob@test.com', 'id': 2, 'validation_rule': 'EMAIL'},
        {'name': 'first_name', 'value': 'joe', 'id': 3, 'validation_rule': 'OPTIONAL'},
        {'name': 'last_name', 'value': 'bob', 'id': 3, 'validation_rule': 'OPTIONAL'},
    ]

    def setUp(self):
        self.form = FormFactory.create()


class TestProcessFormSubmission(FormSubmissionTestCase):
    def setUp(self):
        super(TestProcessFormSubmission, self).setUp()

        FormFieldFactory(form=self.form, template__name="name")
        FormFieldFactory(form=self.form, template__name="email")
        FormFieldFactory(form=self.form, template__name="first_name")
        FormFieldFactory(form=self.form, template__name="last_name")

        self.factory = RequestFactory()
        self.data_obj = {'raw_data': self.default_raw_data}

    @mock.patch('apollo.signals.form_submitted.send')
    @mock.patch('apollo.signals.form_submission_cleaned.send')
    def test_fires_form_submitted_signal(self, mock_form_submission_cleaned_signal, mock_form_submitted_signal):
        request = self.factory.post('/api/v1/submissions')
        self.form.process_submission(self.data_obj, request=request)

        self.assertIsNone(
            mock_form_submitted_signal.assert_called_once_with(
                sender=Form,
                form_id=self.form.id,
                raw_data=self.data_obj,
                request=request
            )
        )

    @mock.patch('apollo.signals.form_submission_cleaned.send')
    def test_fires_form_submission_cleaned_signal(self, mock_form_submission_cleaned_signal):
        request = self.factory.post('/api/v1/submissions')
        submission = self.form.process_submission(self.data_obj, request=request)

        self.assertIsNone(
            mock_form_submission_cleaned_signal.assert_called_once_with(
                sender=Form,
                form_id=self.form.id,
                submission_id=submission.id,
                cleaned_data=submission.cleaned_data,
                request=request
            )
        )

    @mock.patch('apollo.models.FormSubmission.clean_data')
    def test_cleans_form_submission(self, mock_clean_data):
        request = self.factory.post('/api/v1/submissions')
        try:
            self.form.process_submission(self.data_obj, request=request)
        except Exception:
            # we expect this to throw errors, since by mocking clean_data we are unable to set the cleaned_data field
            # on the submission
            pass

        self.assertEqual(mock_clean_data.call_count, 1)


class TestFormSubmissionLabel(FormSubmissionTestCase):
    def test_label_is_email(self):
        FormFieldFactory(
            form=self.form,
            template=FormFieldTemplateFactory(name='name', is_submission_label=False)
        )
        FormFieldFactory(
            form=self.form,
            template=FormFieldTemplateFactory(name='email', is_submission_label=True)
        )

        submission = FormSubmissionFactory(form=self.form, raw_data=self.default_raw_data)

        submission.clean_data()

        self.assertEqual(submission.cleaned_data['email'], submission.label)

    def test_no_labelling_field_set(self):
        FormFieldFactory(
            form=self.form,
            template=FormFieldTemplateFactory(name='name', is_submission_label=False)
        )
        FormFieldFactory(
            form=self.form,
            template=FormFieldTemplateFactory(name='email', is_submission_label=False)
        )

        submission = FormSubmissionFactory(form=self.form, raw_data=self.default_raw_data)

        submission.clean_data()

        self.assertIsNone(submission.label)


class TestFormSubmissionCleanData(FormSubmissionTestCase):
    def setUp(self):
        super(TestFormSubmissionCleanData, self).setUp()

        self.field_name = FormFieldFactory(
            form=self.form,
            default_value='Bob',
            template=FormFieldTemplateFactory(name='name')
        )
        self.field_email = FormFieldFactory(
            form=self.form,
            template=FormFieldTemplateFactory(name='email')
        )

    def test_applies_cleaning_to_raw_data(self):
        submission = FormSubmissionFactory(form=self.form, raw_data={'fields': self.default_raw_data})

        submission.clean_data()

        self.assertDictEqual(
            submission.cleaned_data,
            {
                'email': 'bob@test.com',
                'name': 'bob'
            }
        )

    def test_applying_cleaning_to_raw_data_as_list_is_equivalent(self):
        self.assertDictEqual(
            FormSubmissionFactory(form=self.form, raw_data=self.default_raw_data).clean_data(),
            FormSubmissionFactory(form=self.form, raw_data={'fields': self.default_raw_data}).clean_data()
        )

    def test_strips_fields_not_in_form(self):
        raw_data = {
            'fields': [
                {'name': 'name', 'value': 'bob', 'id': 1, 'validation_rule': 'REQUIRED'},
                {'name': 'email', 'value': 'bob@test.com', 'id': 2, 'validation_rule': 'EMAIL'},
                {'name': 'phone_number', 'value': '1234567890', 'id': 3, 'validation_rule': 'REQUIRED'}
            ]
        }

        submission = FormSubmissionFactory(form=self.form, raw_data=raw_data)

        submission.clean_data()

        self.assertDictEqual(
            submission.cleaned_data,
            {
                'email': 'bob@test.com',
                'name': 'bob'
            }
        )

    def test_adds_fields_not_submitted_and_with_default_values_to_form(self):
        raw_data = [
            {'name': 'email', 'value': 'bob@test.com', 'id': 2, 'validation_rule': 'EMAIL'}
        ]

        submission = FormSubmissionFactory(form=self.form, raw_data=raw_data)

        submission.clean_data()

        self.assertDictEqual(
            submission.cleaned_data,
            {
                'email': 'bob@test.com',
                'name': self.field_name.default_value
            }
        )

    def test_uses_default_field_value_when_field_has_no_value(self):
        raw_data = {
            'fields': [
                {'name': 'name', 'id': 1, 'validation_rule': 'REQUIRED'},
                {'name': 'email', 'value': 'bob@test.com', 'id': 2, 'validation_rule': 'EMAIL'},
            ]
        }

        submission = FormSubmissionFactory(form=self.form, raw_data=raw_data)

        submission.clean_data()

        self.assertDictEqual(
            submission.cleaned_data,
            {
                'email': 'bob@test.com',
                'name': self.field_name.default_value
            }
        )

    @mock.patch('apollo.forms.SubmissionForm.is_valid', return_value=False)
    def test_raises_validation_error_if_validation_fails(self, mock_is_valid):
        raw_data = [
            {'name': 'email', 'value': 'bob@test.com', 'id': 2, 'validation_rule': 'EMAIL'}
        ]

        submission = FormSubmissionFactory(form=self.form, raw_data=raw_data)

        self.assertRaises(ValidationError, submission.clean_data)
        self.assertFalse(submission.is_valid)

    def test_sets_is_valid_field_on_submission(self):
        raw_data = [
            {'name': 'email', 'value': 'bob@test.com', 'id': 2, 'validation_rule': 'EMAIL'}
        ]

        submission = FormSubmissionFactory(form=self.form, raw_data=raw_data)

        submission.clean_data()

        self.assertTrue(submission.is_valid)


#####-----< Webhooks >----#####
class TestExternalWebhookSendData(TestCase):
    def setUp(self):
        self.form = FormFactory(allow_webhooks=True)
        self.hook = ExternalWebhookFactory(
            form=self.form,
            url='https://test.com/',
            for_event=ExternalWebhook.EVENT_SUBMISSION_CREATED
        )
        self.data = {'first_name': 'Bobby', 'last_name': 'S', 'email': 'test@test.com'}

    @responses.activate
    def test_sends_passed_data_to_webhook_url(self):
        responses.add(responses.POST, self.hook.url, status=200)

        self.hook.send_data(self.data)

        self.assertEqual(responses.calls[0].request.url, self.hook.url)
        self.assertEqual(responses.calls[0].request.body.decode('utf-8'), json.dumps(self.data))

    @responses.activate
    @mock.patch('apollo.signals.external_webhook_error.send')
    def test_request_returns_bad_status_code_sends_webhook_error_signal(self, mock_external_webhook_error_signal):
        responses.add(responses.POST, self.hook.url, status=500, body='Server Error')

        self.hook.send_data(self.data)

        self.assertIsNone(
            mock_external_webhook_error_signal.assert_called_once_with(
                ExternalWebhook,
                error='500 Server Error: Internal Server Error for url: https://test.com/',
                url=self.hook.url,
                data=self.data
            )
        )

    @responses.activate
    @mock.patch('apollo.signals.external_webhook_error.send')
    def test_request_raises_http_error_sends_webhook_error_signal(self, mock_external_webhook_error_signal):
        responses.add(responses.POST, self.hook.url, status=200, body=requests.ConnectionError('Error Connecting'))

        self.hook.send_data(self.data)

        self.assertIsNone(
            mock_external_webhook_error_signal.assert_called_once_with(
                ExternalWebhook,
                error='Error Connecting',
                url=self.hook.url,
                data=self.data
            )
        )


from django.views.decorators.clickjacking import xframe_options_exempt
from django.conf.urls import url, include
from django.contrib.staticfiles.urls import staticfiles_urlpatterns
from django.contrib.admin.views.decorators import staff_member_required
from django.conf import settings
from .settings import apollo_settings
from .views import FormManager, DownloadFormSubmissions, NoScriptForm
from .api import urls as api_urls
import os


form_builder_view = FormManager.as_view()
if apollo_settings.BUILDER_REQUIRES_IS_STAFF:
    form_builder_view = staff_member_required(form_builder_view)

download_submissions_view = DownloadFormSubmissions.as_view()
if apollo_settings.DOWNLOAD_REQUIRES_IS_STAFF:
    download_submissions_view = staff_member_required(download_submissions_view)

urlpatterns = [
    url(r'builder/', form_builder_view),
    url(r'download/', download_submissions_view),
    url(r'render/', xframe_options_exempt(NoScriptForm.as_view()), name='render_noscript'),
    url(r'%s/' % os.path.join(apollo_settings.API_ROOT, apollo_settings.API_VERSION), include(api_urls))
]

if settings.DEBUG:
    urlpatterns += staticfiles_urlpatterns()


from django.shortcuts import render, redirect
from django.views.generic import TemplateView
from django.http import HttpResponse
from django.utils.encoding import smart_str
from django.urls import reverse
from django.core.files import File
from django.core.exceptions import ValidationError
from django.core.files.temp import NamedTemporaryFile
from apollo.settings import apollo_settings
from apollo.forms import *
from apollo.models import Form
import logging
import csv
import os
import pdb

logger = logging.getLogger(__name__)


class FormManager(TemplateView):
    def get(self, request, *args, **kwargs):
        return render(request, 'forms_manager.html', {
            'api_root': '/' + os.path.join(apollo_settings.API_ROOT, apollo_settings.API_VERSION),
            'root_prefix': apollo_settings.BUILDER_ROOT_PREFIX,
            'themes_url': apollo_settings.ICARUS_THEMES_URL,
        })


class DownloadFormSubmissions(TemplateView):
    def get(self, request, *args, **kwargs):
        form_id = request.GET.get('form_id')

        form = Form.objects.get(id=form_id)
        fields = form.fields.select_related('template').all().order_by('template__name')
        submissions = form.submissions.all()

        # write the submissions to a CSV
        f = NamedTemporaryFile(suffix='.csv')

        logger.debug('writing submission data to tmp file %s' % f.name)

        with open(f.name, 'w') as csvfile:
            writer = csv.DictWriter(csvfile, [_.name for _ in fields])

            writer.writeheader()

            for s in submissions:
                writer.writerow(s.cleaned_data)

        response = HttpResponse(f.read(), content_type='application/force-download')
        response['Content-Disposition'] = 'attachment; filename=%s.csv' % smart_str('%s_submissions' % form.name.replace(' ', '_'))

        return response


class NoScriptForm(TemplateView):
    """ the view for rendering a given form for the case when javascript isn't available
    NOTE: This is wrapped in urls.py by xframe_options_exempt, so if for some reason you want to limit requests to this view,
    or deny them altogether, you'll need to configure the X-Frame-Options header in your webserver directly
    """
    def get(self, request, *args, **kwargs):
        form_id = request.GET.get('form_id')

        form_obj = Form.objects.get(id=form_id)

        return render(request, 'noscript.html', {
            'form_id': form_id,
            'form': SubmissionForm(apollo_fields=form_obj.fields.all()),
            'themes_url': apollo_settings.ICARUS_THEMES_URL
        })

    def post(self, request):
        """ we will only implement this if we observe the no-script form seeing heavy usage, since this will require
        a thoughtful refactor
        """
        data = request.POST
        data_obj = {'raw_data': [{'name': k, 'value': v} for k,v in data.items()]}
        form_id = data['form_id']

        form_obj = Form.objects.get(id=form_id)

        try:
            submission = form_obj.process_submission(data_obj, request=request)
        except ValidationError as exc:
            return render(request, 'noscript.html', {
                'form_id': form_id,
                'form': SubmissionForm(apollo_fields=form_obj.fields.all()),
                'themes_url': apollo_settings.ICARUS_THEMES_URL,
                'error': str(exc)
            })

        return redirect('{}?form_id={}'.format(reverse('render_noscript'), form_id))




from django import forms


class SubmissionForm(forms.Form):
    def __init__(self, *args, **kwargs):
        # apollo fields are the Apollo FormField objects defined in an Apollo Form object
        apollo_fields = kwargs.pop('apollo_fields', [])
        super(SubmissionForm, self).__init__(*args, **kwargs)

        for f in apollo_fields:
            field_kwargs = dict(
                required=f.is_required,
                label=f.label,
                initial=f.default_value
            )

            widget = self.get_widget_for_apollo_field(f)
            if widget:
                field_kwargs['widget'] = widget(attrs=self.get_widget_attrs_for_apollo_field(f))

            if f.validation_rule == 'EMAIL':
                self.fields[f.name] = forms.EmailField(**field_kwargs)
            else:
                self.fields[f.name] = forms.CharField(**field_kwargs)

    def get_widget_for_apollo_field(self, af):
        widget_map = {
            'TEXT': forms.TextInput,
            'TEXTAREA': forms.Textarea,
            'SELECT': forms.Select,
            'CHECKBOX': forms.CheckboxInput,
            'RADIO': forms.RadioSelect
        }

        # hidden is a bit of a special case, since the field may be hidden as a result of being input type = Hiddden OR
        # by virtue of having it's is_visible field = False
        try:
            return forms.HiddenInput if not af.is_visible else widget_map[af.input_type]
        except KeyError:
            return None

    def get_widget_attrs_for_apollo_field(self, af):
        attrs = {}

        if af.placeholder:
            attrs['placeholder'] = af.placeholder

        return attrs



from __future__ import unicode_literals

from django.db import models
from django.contrib.postgres.fields import ArrayField, JSONField
from django.core.exceptions import ValidationError
from django.conf import settings
from apollo import forms, signals
from apollo.settings import apollo_settings
from apollo.lib import emails, recaptcha
from apollo.exceptions import RecaptchaVerificationException
import pdb
import requests
import copy
import json
import logging

logger = logging.getLogger(__name__)


#####-----< Layout >----#####
class Layout(models.Model):
    """ a layout controls the positioning and relative sizing of fields in a form."""
    name = models.CharField(max_length=150, unique=True)
    max_width = models.IntegerField(null=True, blank=True, default=None)

    def __str__(self):
        return """%s (%s)""" % (
            self.name, self.id
        )


class LayoutField(models.Model):
    """ the positioning of a field in the forms' layout """
    layout = models.ForeignKey(Layout, related_name='blocks')

    # width and height are in pct relative
    desktop_width = models.DecimalField(max_digits=6, decimal_places=3)
    desktop_height = models.DecimalField(max_digits=6, decimal_places=3)
    desktop_top_left = ArrayField(models.DecimalField(max_digits=6, decimal_places=3), size=2)

    mobile_width = models.DecimalField(max_digits=6, decimal_places=3)
    mobile_height = models.DecimalField(max_digits=6, decimal_places=3)
    mobile_top_left = ArrayField(models.DecimalField(max_digits=6, decimal_places=3), size=2)

    def __str__(self):
        return """%s x %s @ %s (%s)""" % (
            self.desktop_width, self.desktop_height, self.desktop_top_left, self.id
        )


#####-----< Forms >----#####
class Form(models.Model):
    name = models.CharField(max_length=200, unique=True)

    redirect_url = models.CharField(max_length=150, null=True, blank=True, default=None)
    success_message = models.TextField(null=True, blank=True, default=None)
    submit_button_text = models.CharField(max_length=100, default='Submit')
    # if set, submissions go here instead of to our form submission endpoint
    submission_url = models.CharField(max_length=200, null=True, blank=True, default=None)
    # if set, these email addresses will be notified on submissions
    submission_contacts = ArrayField(models.CharField(max_length=400), null=True, blank=True, default=None)
    # if true, then we allow external hooks to listen for events occuring on this form (submissions, validation, etc.)
    allow_webhooks = models.BooleanField(default=False)
    # if true, forms will use recaptcha v3 for verification
    include_recaptcha = models.BooleanField(default=False)

    layout = models.ForeignKey(Layout, related_name='forms', null=True, blank=True, default=None)

    @classmethod
    def create_from_json(cls, form_data):
        """ instantiates a form instance and its associated fields from its serialized variation.

        :param: json_form `dict` parsed input form data (including field information)
        :return: `Form` the instantiated form instance
        :raises: `ValidationError` if the serializer is given invalid data
        """
        from apollo.api import serializers

        existing_form_field_template_names = set(FormFieldTemplate.objects.values_list('name', flat=True))

        form_serializer = serializers.FormSerializer(data=form_data)
        logger.debug(f'form data is {form_data}')
        if form_serializer.is_valid():
            # we need to create field templates for all fields don't currently exist for the form
            for form_field in form_serializer.validated_data['fields']:
                if form_field['name'] not in existing_form_field_template_names:
                    field_serializer = serializers.FormFieldTemplateSerializer(data=form_field)
                    if field_serializer.is_valid():
                        field_serializer.save()

            # then save the form and all fields to the DB
            return form_serializer.save()
        else:
            raise ValidationError(f'form serializer failed to deserialize with error(s) {form_serializer.errors}')

    def clone(self):
        """ clone this form to a new form. In addition to the basic fields of this model, this has the effect of copying:
        - associated form fields, creating new copies of fields for the new form

        :return: `Form` a clone of this form
        """
        form_copy = copy.deepcopy(self)

        # make a copy of the base form
        form_copy.pk = None
        form_copy.name = self.name + ' Copy'
        form_copy.save()

        # copy the fields
        for field in self.fields.all():
            field.pk = None
            field.form = form_copy
            field.save()

        return form_copy

    def process_submission(self, cleaned_data, request=None):
        signals.form_submitted.send(
            sender=self.__class__,
            form_id=self.id,
            raw_data=cleaned_data,
            request=request
        )

        submission = FormSubmission(raw_data=cleaned_data['raw_data'], form=self)
        error_msg = False

        try:
            submission.clean_data()
        except ValueError as exc:
            error_msg = 'got error cleaning submission data'
        except ValidationError as exc:
            error_msg = 'got error validating submission data'
        finally:
            submission.save()

        # if the submission should be recaptcha protected, verify the recaptcha
        if self.include_recaptcha:
            logger.debug('form %s requests recaptcha checks, checking token now' % self.id)

            try:
                recaptcha_valid = recaptcha.verify_token(cleaned_data.get('recaptcha_token'), recaptcha_secret_key=apollo_settings.RECAPTCHA_SECRET_KEY)
                if not recaptcha_valid:
                    error_msg = 'recaptcha token is invalid'
            except RecaptchaVerificationException:
                error_msg = 'recaptcha verification failed with an error'

        if error_msg:
            signals.form_submission_error.send(
                sender=self.__class__,
                form_id=submission.form_id,
                raw_data=submission.raw_data,
                error=error_msg,
                request=request
            )

            raise ValidationError(error_msg)

        if self.submission_contacts and apollo_settings.SUBMISSION_EMAIL_FROM:
            emails.send(
                from_email=apollo_settings.SUBMISSION_EMAIL_FROM,
                to_emails=self.submission_contacts,
                subject='Apollo: New Form Submission',
                content='new form submission, data = %s' % str(submission.cleaned_data)
            )

        signals.form_submission_cleaned.send(
            sender=self.__class__,
            form_id=submission.form_id,
            submission_id=submission.id,
            cleaned_data=submission.cleaned_data,
            request=request
        )

        return submission

    def __str__(self):
        return """%s (%s)""" % (self.name, self.id)

    class Meta:
        ordering = ['name']


class FormFieldTemplate(models.Model):
    FIELD_TYPES = [
        ('TEXT', 'TEXT'),
        ('HIDDEN', 'HIDDEN'),
        ('TEXTAREA', 'TEXTAREA'),
        ('SELECT', 'SELECT'),
        ('CHECKBOX', 'CHECKBOX'),
        ('RICHTEXT', 'RICHTEXT'),
        ('RADIO', 'RADIO'),
    ]

    RULE_OPTIONAL = 'OPTIONAL'
    RULE_REQUIRED = 'REQUIRED'
    RULE_EMAIL = 'EMAIL'
    RULE_PHONE_NUMBER = 'PHONE_NUMBER'
    VALIDATION_RULES = [(x, x) for x in [RULE_OPTIONAL, RULE_REQUIRED, RULE_EMAIL, RULE_PHONE_NUMBER]]

    LABEL_TOP = 'TOP'
    LABEL_RIGHT = 'RIGHT'
    LABEL_BOTTOM = 'BOTTOM'
    LABEL_LEFT = 'LEFT'
    LABEL_POSITIONS = [
        (LABEL_TOP, LABEL_TOP),
        (LABEL_RIGHT, LABEL_RIGHT),
        (LABEL_BOTTOM, LABEL_BOTTOM),
        (LABEL_LEFT, LABEL_LEFT),
    ]

    # If True, it means that the value of this field in a submission should be used as the primary label for
    # for the submission
    is_submission_label = models.BooleanField(default=False)

    # Core Fields (Not Overriddeable on FormField instances)
    name = models.CharField(max_length=100, unique=True)
    input_type = models.CharField(max_length=50, choices=FIELD_TYPES, default='text')
    is_visible = models.BooleanField(default=True)

    label = models.CharField(max_length=100, null=True, blank=True, default=None)
    label_position = models.CharField(max_length=50, choices=LABEL_POSITIONS, default=LABEL_TOP)
    placeholder = models.CharField(max_length=150, null=True, blank=True, default=None)
    default_value = models.CharField(max_length=300, null=True, blank=True, default=None)

    value_choices = ArrayField(
        ArrayField(models.CharField(max_length=100), size=2),
        null=True,
        blank=True,
        default=None
    )
    validation_rule = models.CharField(max_length=50, null=True, blank=True, default=RULE_REQUIRED, choices=VALIDATION_RULES)

    def __str__(self):
        return """%s // %s (%s)""" % (self.name, self.input_type, self.id)


class FormField(models.Model):
    """ a form field is a concrete instance of a field on a form. It must be separate from a FormFieldTemplate
    because while the template governs the common configuration of a field, there are certain properties which
    might need overriding on a per-form basis (designated by a leyou ading _)
    """
    form = models.ForeignKey(Form, related_name='fields')
    template = models.ForeignKey(FormFieldTemplate, related_name='instances')
    layout = models.ForeignKey(LayoutField, null=True, blank=True, default=None, on_delete=models.SET_NULL)

    label = models.CharField(max_length=100, null=True, blank=True, default=None)
    label_position = models.CharField(max_length=50, null=True, blank=True, default=FormFieldTemplate.LABEL_TOP, choices=FormFieldTemplate.LABEL_POSITIONS)
    default_value = models.CharField(max_length=300, null=True, blank=True, default=None)
    placeholder = models.CharField(max_length=150, null=True, blank=True, default=None)

    index = models.IntegerField(default=0)

    value_choices = ArrayField(
        ArrayField(models.CharField(max_length=100), size=2),
        null=True,
        blank=True,
        default=None
    )
    validation_rule = models.CharField(max_length=50, null=True, blank=True, default=FormFieldTemplate.RULE_REQUIRED, choices=FormFieldTemplate.VALIDATION_RULES)

    @property
    def name(self):
        return self.template.name

    @property
    def input_type(self):
        return self.template.input_type

    @property
    def is_visible(self):
        return self.template.is_visible and self.template.input_type != 'HIDDEN'

    @property
    def is_submission_label(self):
        return self.template.is_submission_label

    @property
    def is_required(self):
        return self.validation_rule != FormFieldTemplate.RULE_OPTIONAL

    def __str__(self):
        return """%s // Form %s (%s)""" % (
            self.name, self.form, self.id
        )


#####-----< Submissions >----#####
class FormSubmission(models.Model):
    form = models.ForeignKey(Form, related_name='submissions')

    raw_data = JSONField()
    cleaned_data = JSONField(blank=True, null=True, default=None)

    is_valid = models.BooleanField(default=True)

    created_time = models.DateTimeField(auto_now_add=True)

    @property
    def label(self):
        """ the label for this submission is governed by the templates of the fields of the form. The first template
        which has `is_submission_label = True` has its field value looked up in the cleaned data and used as the submissions'
        label
        """
        labelling_field = self.form.fields.filter(template__is_submission_label=True).first()

        if not labelling_field or not self.cleaned_data:
            return None

        return self.cleaned_data[labelling_field.name]

    def clean_data(self):
        """ raw_data => cleaned_data. Cleaning is the process of:
        - removing fields from the raw data any fields which don't exist on the form being submitted
        - converting the raw_data, which will be a list of field dictionaries containing data submitted by the client,
        into a dictionary of field name -> field value
        - validating the data
        :return: `dict` of cleaned, validated data
        :raises: ValidationError if the submission fails to validate
        :raises: ValueError if data isn't available to clean / validate
        """
        # raw_data can be of form {'fields': [<field data>]} or [<field data]
        try:
            fields_list = self.raw_data['fields']
        except TypeError:
            fields_list = self.raw_data

        if not fields_list:
            raise ValueError('no data was found in the raw_data')

        named_data = {}

        # transform the fields list into a dictionary, indexed by field names
        fields_dict = {f['name']: f for f in fields_list}

        stored_defaults = dict(self.form.fields.all().values_list('template__name', 'default_value'))

        for fname in stored_defaults:
            if stored_defaults[fname] or fname in fields_dict:
                try:
                    named_data[fname] = fields_dict[fname]['value']
                except KeyError:
                    named_data[fname] = stored_defaults[fname]

        sform = forms.SubmissionForm(apollo_fields=self.form.fields.all(), data=named_data)

        self.is_valid = sform.is_valid()

        if self.is_valid:
            self.cleaned_data = sform.cleaned_data
        else:
            raise ValidationError(json.dumps(sform.errors.as_json()))

        return self.cleaned_data

    def __str__(self):
        return """Submission for %s @ %s (%s)""" % (
            self.form, self.created_time, self.id
        )

    class Meta:
        ordering = ['-created_time', 'is_valid']
        permissions = (
            ("can_view_submissions", "Can View Form Submissions"),
            ("can_download_submissions", "Can Download Form Submissions"),
        )


#####-----< API >----#####
class APIUser(models.Model):
    """ an APIUser is the only type we allow to authenticate with Token auth (as opposed to session auth) """
    auth_user = models.ForeignKey(settings.AUTH_USER_MODEL, related_name='apis')
    service_name = models.CharField(max_length=200)

    def __str__(self):
        return """%s -- API = %s (%s)""" % (
            self.auth_user, self.service_name, self.id
        )


class ExternalWebhook(models.Model):
    """ an external webhook is a URL to which we POST when an event occurs against a Form (or child object) """
    # the allowed events define the possible events a webhook might be registered to listen on (one per hook)
    EVENT_SUBMISSION_CREATED = "new form submission"
    ALLOWED_EVENTS = (
        (EVENT_SUBMISSION_CREATED, EVENT_SUBMISSION_CREATED),
    )

    form = models.ForeignKey(Form, related_name='external_hooks')

    url = models.CharField(max_length=255)
    for_event = models.CharField(max_length=100, choices=ALLOWED_EVENTS)
    # the name of the service registering a hook
    service = models.CharField(max_length=100)

    def send_data(self, data):
        """ send the given data to this webhooks' URL """
        err = resp = None

        try:
            resp = requests.post(self.url, json=data)
        except Exception as exc:
            # only request exceptions can occur here, so we blanket except
            err = str(exc)

        if resp is not None and not resp.ok:
            try:
                resp.raise_for_status()
            except requests.HTTPError as exc:
                err = str(exc)

        if err:
            signals.external_webhook_error.send(
                self.__class__,
                error=err,
                url=self.url,
                data=data
            )

        return resp.status_code if resp else None

    def __str__(self):
        return """%s: Send %s -> %s (%s)""" % (
            self.form.name, self.for_event, self.service.upper(), self.id
        )



# -*- coding: utf-8 -*-
# Generated by Django 1.11.4 on 2018-02-12 17:22
from __future__ import unicode_literals

from django.conf import settings
import django.contrib.postgres.fields
import django.contrib.postgres.fields.jsonb
from django.db import migrations, models
import django.db.models.deletion


class Migration(migrations.Migration):

    initial = True

    dependencies = [
        migrations.swappable_dependency(settings.AUTH_USER_MODEL),
    ]

    operations = [
        migrations.CreateModel(
            name='APIUser',
            fields=[
                ('id', models.AutoField(auto_created=True, primary_key=True, serialize=False, verbose_name='ID')),
                ('service_name', models.CharField(max_length=200)),
                ('auth_user', models.ForeignKey(on_delete=django.db.models.deletion.CASCADE, related_name='apis', to=settings.AUTH_USER_MODEL)),
            ],
        ),
        migrations.CreateModel(
            name='Form',
            fields=[
                ('id', models.AutoField(auto_created=True, primary_key=True, serialize=False, verbose_name='ID')),
                ('name', models.CharField(max_length=200, unique=True)),
                ('redirect_url', models.CharField(blank=True, default=None, max_length=150, null=True)),
                ('success_message', models.TextField(blank=True, default=None, null=True)),
                ('submit_button_text', models.CharField(default='Submit', max_length=100)),
                ('submission_url', models.CharField(blank=True, default=None, max_length=200, null=True)),
                ('submission_contacts', django.contrib.postgres.fields.ArrayField(base_field=models.CharField(max_length=400), blank=True, default=None, null=True, size=None)),
            ],
            options={
                'ordering': ['name'],
            },
        ),
        migrations.CreateModel(
            name='FormField',
            fields=[
                ('id', models.AutoField(auto_created=True, primary_key=True, serialize=False, verbose_name='ID')),
                ('label', models.CharField(blank=True, default=None, max_length=100, null=True)),
                ('label_position', models.CharField(blank=True, choices=[('TOP', 'TOP'), ('RIGHT', 'RIGHT'), ('BOTTOM', 'BOTTOM'), ('LEFT', 'LEFT')], default='TOP', max_length=50, null=True)),
                ('default_value', models.CharField(blank=True, default=None, max_length=300, null=True)),
                ('placeholder', models.CharField(blank=True, default=None, max_length=150, null=True)),
                ('index', models.IntegerField(default=0)),
                ('value_choices', django.contrib.postgres.fields.ArrayField(base_field=django.contrib.postgres.fields.ArrayField(base_field=models.CharField(max_length=100), size=2), blank=True, default=None, null=True, size=None)),
                ('validation_rule', models.CharField(blank=True, choices=[('OPTIONAL', 'OPTIONAL'), ('REQUIRED', 'REQUIRED'), ('EMAIL', 'EMAIL')], default='REQUIRED', max_length=50, null=True)),
                ('form', models.ForeignKey(on_delete=django.db.models.deletion.CASCADE, related_name='fields', to='apollo.Form')),
            ],
        ),
        migrations.CreateModel(
            name='FormFieldTemplate',
            fields=[
                ('id', models.AutoField(auto_created=True, primary_key=True, serialize=False, verbose_name='ID')),
                ('is_submission_label', models.BooleanField(default=False)),
                ('name', models.CharField(max_length=100, unique=True)),
                ('input_type', models.CharField(choices=[('TEXT', 'TEXT'), ('HIDDEN', 'HIDDEN'), ('TEXTAREA', 'TEXTAREA'), ('SELECT', 'SELECT'), ('CHECKBOX', 'CHECKBOX'), ('RICHTEXT', 'RICHTEXT'), ('RADIO', 'RADIO')], default='text', max_length=50)),
                ('is_visible', models.BooleanField(default=True)),
                ('label', models.CharField(blank=True, default=None, max_length=100, null=True)),
                ('label_position', models.CharField(choices=[('TOP', 'TOP'), ('RIGHT', 'RIGHT'), ('BOTTOM', 'BOTTOM'), ('LEFT', 'LEFT')], default='TOP', max_length=50)),
                ('placeholder', models.CharField(blank=True, default=None, max_length=150, null=True)),
                ('default_value', models.CharField(blank=True, default=None, max_length=300, null=True)),
                ('value_choices', django.contrib.postgres.fields.ArrayField(base_field=django.contrib.postgres.fields.ArrayField(base_field=models.CharField(max_length=100), size=2), blank=True, default=None, null=True, size=None)),
                ('validation_rule', models.CharField(blank=True, choices=[('OPTIONAL', 'OPTIONAL'), ('REQUIRED', 'REQUIRED'), ('EMAIL', 'EMAIL')], default='REQUIRED', max_length=50, null=True)),
            ],
        ),
        migrations.CreateModel(
            name='FormSubmission',
            fields=[
                ('id', models.AutoField(auto_created=True, primary_key=True, serialize=False, verbose_name='ID')),
                ('raw_data', django.contrib.postgres.fields.jsonb.JSONField()),
                ('cleaned_data', django.contrib.postgres.fields.jsonb.JSONField(blank=True, default=None, null=True)),
                ('is_valid', models.BooleanField(default=True)),
                ('created_time', models.DateTimeField(auto_now_add=True)),
                ('form', models.ForeignKey(on_delete=django.db.models.deletion.CASCADE, related_name='submissions', to='apollo.Form')),
            ],
            options={
                'ordering': ['-created_time', 'is_valid'],
                'permissions': (('can_view_submissions', 'Can View Form Submissions'), ('can_download_submissions', 'Can Download Form Submissions')),
            },
        ),
        migrations.CreateModel(
            name='Layout',
            fields=[
                ('id', models.AutoField(auto_created=True, primary_key=True, serialize=False, verbose_name='ID')),
                ('name', models.CharField(max_length=150, unique=True)),
                ('max_width', models.IntegerField(blank=True, default=None, null=True)),
            ],
        ),
        migrations.CreateModel(
            name='LayoutField',
            fields=[
                ('id', models.AutoField(auto_created=True, primary_key=True, serialize=False, verbose_name='ID')),
                ('desktop_width', models.DecimalField(decimal_places=3, max_digits=6)),
                ('desktop_height', models.DecimalField(decimal_places=3, max_digits=6)),
                ('desktop_top_left', django.contrib.postgres.fields.ArrayField(base_field=models.DecimalField(decimal_places=3, max_digits=6), size=2)),
                ('mobile_width', models.DecimalField(decimal_places=3, max_digits=6)),
                ('mobile_height', models.DecimalField(decimal_places=3, max_digits=6)),
                ('mobile_top_left', django.contrib.postgres.fields.ArrayField(base_field=models.DecimalField(decimal_places=3, max_digits=6), size=2)),
                ('layout', models.ForeignKey(on_delete=django.db.models.deletion.CASCADE, related_name='blocks', to='apollo.Layout')),
            ],
        ),
        migrations.AddField(
            model_name='formfield',
            name='layout',
            field=models.ForeignKey(blank=True, default=None, null=True, on_delete=django.db.models.deletion.SET_NULL, to='apollo.LayoutField'),
        ),
        migrations.AddField(
            model_name='formfield',
            name='template',
            field=models.ForeignKey(on_delete=django.db.models.deletion.CASCADE, related_name='instances', to='apollo.FormFieldTemplate'),
        ),
        migrations.AddField(
            model_name='form',
            name='layout',
            field=models.ForeignKey(blank=True, default=None, null=True, on_delete=django.db.models.deletion.CASCADE, related_name='forms', to='apollo.Layout'),
        ),
    ]



"""
Settings for Apollo are all namespaced in the APOLLO setting.
For example your project's `settings.py` file might look like this:
APOLLO = {
    'ROOT_PATH': '/'
}
This module provides the `apollo_settings` object, that is used to access
Graphene settings, checking for user settings first, then falling
back to the defaults.
"""
from __future__ import unicode_literals

from django.conf import settings
from django.test.signals import setting_changed

try:
    import importlib  # Available in Python 3.1+
except ImportError:
    from django.utils import importlib  # Will be removed in Django 1.9


# Copied shamelessly from Graphene which copies shamelessly from Django REST Framework

DEFAULTS = {
    'API_ROOT': 'api', # the path to the REST api
    'API_VERSION': 'v1', # the REST api version
    'BUILDER_ROOT_PREFIX': '', # the path at which the apollo URL's are mounted. This should be set if the path is anything but the root
    'BUILDER_REQUIRES_IS_STAFF': True, # if True, then usage of the form builder requires is_staff privileges
    'DOWNLOAD_REQUIRES_IS_STAFF': True, # if True, then downloading submissions for a form requires is_staff privileges
    'ICARUS_THEMES_URL': '//static.forthepeople.com/engineering/icarus/v1.11-latest/themes', # location of the icarus themes directory
    'SUBMISSION_EMAIL_FROM': None, # email address to send submission notifications from
    'EXTERNAL_HOOKS_SIGNAL_ONLY': True, # if False, then we won't just send a signal notifying listeners that a webhook has been tripped, we'll also make the request to the hook
    'RECAPTCHA_SITE_KEY': None, # the site key for recaptcha v3 (only required if you want to include recaptcha in forms)
    'RECAPTCHA_SECRET_KEY': None # the recaptcha v3 secret key (only required if you want to include recaptcha in forms)
}


class ApolloSettings(object):
    """
    A settings object, that allows API settings to be accessed as properties.
    For example:
        from apollo.settings import settings
        print(settings.ROOT_PATH)
    Any setting with string import paths will be automatically resolved
    and return the class, rather than the string literal.
    """

    def __init__(self, user_settings=None, defaults=None):
        if user_settings:
            self._user_settings = user_settings
        self.defaults = defaults or DEFAULTS

    @property
    def user_settings(self):
        if not hasattr(self, '_user_settings'):
            self._user_settings = getattr(settings, 'APOLLO', {})
        return self._user_settings

    def __getattr__(self, attr):
        if attr not in self.defaults:
            raise AttributeError("Invalid Apollo setting: '%s'" % attr)

        try:
            # Check if present in user settings
            val = self.user_settings[attr]
        except KeyError:
            # Fall back to defaults
            val = self.defaults[attr]

        # Cache the result
        setattr(self, attr, val)
        return val


apollo_settings = ApolloSettings(None, DEFAULTS)


def reload_apollo_settings(*args, **kwargs):
    global apollo_settings
    setting, value = kwargs['setting'], kwargs['value']
    if setting == 'APOLLO':
        apollo_settings = ApolloSettings(value, DEFAULTS)


setting_changed.connect(reload_apollo_settings)



from django.test import TestCase
from apollo.factories import FormFieldFactory, FormFactory, FormSubmissionFactory, ExternalWebhookFactory
from apollo import signals, models, settings
import requests
import responses
import mock


class SignalTestCase(TestCase):
    pass


class TestFireExternalWebhooksSignal(SignalTestCase):
    def setUp(self):
        super(TestFireExternalWebhooksSignal, self).setUp()

        self.form = FormFactory(allow_webhooks=True)
        self.hook = ExternalWebhookFactory(form=self.form, url='https://test.com', for_event=models.ExternalWebhook.EVENT_SUBMISSION_CREATED)
        self.submission = FormSubmissionFactory(
            form=self.form,
            raw_data={},
            cleaned_data={'first_name': 'Bobby', 'last_name': 'S', 'email': 'test@test.com'},
            is_valid=True
        )

    @mock.patch('apollo.models.ExternalWebhook.send_data')
    def test_calls_webhook_send_data_with_submission_data(self, mock_send_data):
        mock_send_data.return_value = responses.Response(status=200, method='POST', url=self.hook.url)

        signals.fire_form_submission_webhooks(None, self.submission.cleaned_data, self.form.id, self.submission.id)

        self.assertIsNone(
            # we could use FormSubmissionSerializer(instance=submission).data here, but I think this is clearer
            mock_send_data.assert_called_once_with({
                'id': self.submission.id,
                'cleaned_data': self.submission.cleaned_data,
                'raw_data': self.submission.raw_data,
                'created_time': self.submission.created_time.strftime('%Y-%m-%dT%H:%M:%S.%fZ'),
                'form': self.form.id,
                'is_valid': True,
                'label': self.submission.label
            })
        )




from django.conf.urls import url, include
from rest_framework import routers
from rest_framework.authtoken import views as token_views
from rest_framework_swagger.views import get_swagger_view
from . import api_views


router = routers.SimpleRouter(trailing_slash=False)
router.register(r'forms', api_views.FormViewSet)
router.register(r'submissions', api_views.FormSubmissionViewSet)
router.register(r'field-templates', api_views.FormFieldTemplateViewSet)
router.register(r'fields', api_views.FormFieldViewSet)
router.register(r'layouts', api_views.LayoutViewSet)
router.register(r'hooks', api_views.ExternalWebhookViewSet)

urlpatterns = router.urls

schema_view = get_swagger_view(title='Apollo API')

urlpatterns += (
    url(r'field-options/', api_views.FormFieldOptionsAPIView.as_view()),
    url(r'auth-token', token_views.obtain_auth_token),
    url(r'docs/', schema_view)
)


import factory
from apollo import models


class FormFactory(factory.django.DjangoModelFactory):
    name = factory.Faker('name')
    redirect_url = factory.Faker('url')
    submission_url = factory.Faker('uri_page')

    class Meta:
        model = models.Form


class FormFieldTemplateFactory(factory.django.DjangoModelFactory):
    name = factory.Faker('name')

    class Meta:
        model = models.FormFieldTemplate


class FormFieldFactory(factory.django.DjangoModelFactory):
    form = factory.SubFactory(FormFactory)
    template = factory.SubFactory(FormFieldTemplateFactory)

    class Meta:
        model = models.FormField


class FormSubmissionFactory(factory.django.DjangoModelFactory):
    form = factory.SubFactory(FormFactory)

    class Meta:
        model = models.FormSubmission


class APIUserFactory(factory.django.DjangoModelFactory):
    service_name = factory.Faker('company')

    class Meta:
        model = models.APIUser


class ExternalWebhookFactory(factory.django.DjangoModelFactory):
    form = factory.SubFactory(FormFactory)

    class Meta:
        model = models.ExternalWebhook